from os import environ, path
import unittest
import random

import fnss

TMP_DIR = environ['test.tmp.dir'] if 'test.tmp.dir' in environ else None

class Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def event_gen(self, threshold, action):
        event_props = {}
        r = random.random()
        event_props['action'] = action[0] if r > threshold else action[1]
        return event_props

    def test_event_schedule_add(self):
        es = fnss.EventSchedule()
        es.add(8, {'add_order': 1}, absolute_time=True)
        es.add(5, {'add_order': 2}, absolute_time=True)
        self.assertEqual(2, es.number_of_events())
        self.assertEqual(5, es[0][0])
        self.assertEqual({'add_order': 2}, es[0][1])
        self.assertEqual(8, es[1][0])
        self.assertEqual({'add_order': 1}, es[1][1])


    def test_event_schedule_pop(self):
        es = fnss.EventSchedule()
        es.add(8, {'add_order': 1}, absolute_time=True)
        es.add(5, {'add_order': 2}, absolute_time=True)
        t0, e0 = es.pop(0)
        t1, e1 = es.pop(0)
        self.assertEqual(5, t0)
        self.assertEqual(8, t1)
        self.assertEqual({'add_order': 2}, e0)
        self.assertEqual({'add_order': 1}, e1)
        self.assertEqual(0, es.number_of_events())


    def test_event_schedule_events_between(self):
        es = fnss.EventSchedule()
        es.add(5, {'event_order': 3}, absolute_time=True)
        es.add(4, {'event_order': 2}, absolute_time=True)
        es.add(3, {'event_order': 1}, absolute_time=True)
        es.add(7, {'event_order': 5}, absolute_time=True)
        es.add(6, {'event_order': 4}, absolute_time=True)
        es.add(8, {'event_order': 6}, absolute_time=True)
        events = es.events_between(5, 7)
        self.assertEqual(2, events.number_of_events())
        self.assertEqual(5, events[0][0])
        self.assertEqual({'event_order': 3}, events[0][1])
        self.assertEqual(6, events[1][0])
        self.assertEqual({'event_order': 4}, events[1][1])


    def test_event_schedule_add_schedule(self):
        es1 = fnss.EventSchedule(t_unit='s')
        es1.add(3, {'event_order': 1}, absolute_time=True)
        es1.add(5, {'event_order': 3}, absolute_time=True)
        es2 = fnss.EventSchedule(t_unit='ms')
        es2.add(4000, {'event_order': 2}, absolute_time=True)
        es2.add(7000, {'event_order': 5}, absolute_time=True)
        es1.add_schedule(es2)
        self.assertEqual(4, len(es1))
        self.assertEqual('s', es1.attrib['t_unit'])
        self.assertEqual(3, es1[0][0])
        self.assertEqual(4, es1[1][0])
        self.assertEqual(5, es1[2][0])
        self.assertEqual(7, es1[3][0])

    def test_event_schedule_add_operator(self):
        es1 = fnss.EventSchedule(t_unit='s')
        es1.add(3, {'event_order': 1}, absolute_time=True)
        es1.add(5, {'event_order': 3}, absolute_time=True)
        es2 = fnss.EventSchedule(t_unit='ms')
        es2.add(4000, {'event_order': 2}, absolute_time=True)
        es2.add(7000, {'event_order': 5}, absolute_time=True)
        es3 = es1 + es2
        self.assertEqual(4, len(es3))
        self.assertEqual('s', es3.attrib['t_unit'])
        self.assertEqual(3, es3[0][0])
        self.assertEqual(4, es3[1][0])
        self.assertEqual(5, es3[2][0])
        self.assertEqual(7, es3[3][0])

    def test_event_schedule_operators(self):
        es = fnss.EventSchedule()
        es.add(5, {'event_order': 3}, absolute_time=True)
        es.add(4, {'event_order': 2}, absolute_time=True)
        es.add(3, {'event_order': 1}, absolute_time=True)
        es.add(7, {'event_order': 5}, absolute_time=True)
        es.add(6, {'event_order': 4}, absolute_time=True)
        es.add(8, {'event_order': 6}, absolute_time=True)
        self.assertEqual(6, len(es))
        self.assertEqual(2, len(es[3:5]))
        del es[0]
        self.assertEqual(5, len(es))
        for ev in es:
            t, _ = ev
            self.assertGreaterEqual(t, es.attrib['t_start'])
            self.assertLessEqual(t, es.attrib['t_end'])


    def test_deterministic_process_event_schedule(self):
        action = ['read_email', 'watch_video']
        schedule = fnss.deterministic_process_event_schedule(20, 0, 80001, 'ms',
                                                             self.event_gen,
                                                             0.5, action=action)
        self.assertIsNotNone(schedule)
        self.assertEqual(4000, len(schedule))
        for time, event in schedule:
            self.assertTrue(event['action'] in action)
            self.assertTrue(time >= 0)
            self.assertTrue(time <= 80001)


    def test_poisson_process_event_schedule(self):
        action = ['read_email', 'watch_video']
        schedule = fnss.poisson_process_event_schedule(15, 0, 8000, 'ms',
                                                       self.event_gen,
                                                       0.5, action=action)
        self.assertIsNotNone(schedule)
        for time, event in schedule:
            self.assertTrue(event['action'] in action)
            self.assertTrue(time >= 0)
            self.assertTrue(time <= 8000)

    @unittest.skipIf(TMP_DIR is None, "Temp folder not present")
    def test_read_write_event_schedule(self):
        action = ['read_email', 'watch_video']
        schedule = fnss.deterministic_process_event_schedule(20, 0, 801, 'ms',
                                                             self.event_gen,
                                                             0.5, action=action)
        time, event = schedule[2]
        tmp_es_file = path.join(TMP_DIR, 'event-schedule.xml')
        fnss.write_event_schedule(schedule, tmp_es_file)
        read_schedule = fnss.read_event_schedule(tmp_es_file)
        self.assertEqual(len(schedule), len(read_schedule))
        read_time, read_event = read_schedule[2]
        self.assertEqual(time, read_time)
        self.assertEqual(event, read_event)

    @unittest.skipIf(TMP_DIR is None, "Temp folder not present")
    def test_read_write_event_schedule_special_type(self):
        schedule = fnss.EventSchedule()
        event = {'tuple_param': (1, 2, 3),
                 'dict_param': {'a': 1, 'b': 2},
                 'list_param':[1, 'hello', 0.3]}
        schedule.add(1, event)
        tmp_es_file = path.join(TMP_DIR, 'event-schedule-special.xml')
        fnss.write_event_schedule(schedule, tmp_es_file)
        read_schedule = fnss.read_event_schedule(tmp_es_file)
        self.assertEqual(len(schedule), len(read_schedule))
        _, read_event = read_schedule[0]
        self.assertEqual(event, read_event)
        self.assertEqual(tuple, type(read_event['tuple_param']))
        self.assertEqual(list, type(read_event['list_param']))
        self.assertEqual(dict, type(read_event['dict_param']))
        self.assertEqual(event['dict_param'], read_event['dict_param'])
        self.assertEqual(event['list_param'], read_event['list_param'])
        self.assertEqual(event['tuple_param'], read_event['tuple_param'])



import unittest

import fnss.util as util


class TestSplitList(unittest.TestCase):

    def test_no_remainder(self):
        l = util.split_list([1, 2, 3, 4, 5, 6], 2)
        self.assertEqual(l, [[1, 2], [3, 4], [5, 6]])

    def test_remainder(self):
        l = util.split_list([1, 2, 3], 2)
        self.assertEqual(l, [[1, 2], [3]])


class TestGeographicalDistance(unittest.TestCase):

    def test_normal_case(self):
        d = util.geographical_distance(-30, -30, 30, 30)
        self.assertGreater(d, 0)

    def test_pole_node(self):
        d = util.geographical_distance(90, 30, 40, 90)
        self.assertGreater(d, 0)



"""Omnet++ adapter

This module contains the code for converting an FNSS topology object into a NED
script to deploy such topology into Omnet++.
"""
from warnings import warn

from fnss.units import time_units, capacity_units


__all__ = ['to_omnetpp']


# Template text rendered by the template engine
__TEMPLATE = r"""// Code generated by Fast Network Simulator Setup (FNSS)
<%
import re
from fnss.units import time_units, capacity_units
from fnss.netconfig.nodeconfig import get_stack, \
                                      get_application_names, \
                                      get_application_properties

# Convert capacity in Mb
if set_capacities:
    capacity_norm = capacity_units[topology.graph['capacity_unit']] / 1000000.0
# Convert delay it in ms
if set_delays:
    delay_norm = time_units[topology.graph['delay_unit']]

# Get topology name
name = topology.name
name = "net" if name == "" else re.sub("[^A-Za-z0-9]", "_", name).strip(" _")

# Get numerical ID of a node
nodes = list(topology.nodes())
node_map = {node: i for i, node in enumerate(nodes)}

%>
// This is the module modelling the nodes of network
module node {
    //TODO: Implement it
}

// Create network
network ${name} {
    connections allowunconnected:
% for u, v in topology.edges():
    <%
    attr_str = ""
    if set_delays:
        delay = delay_norm * topology.adj[u][v]['delay']
        attr_str += " delay=%sms;" % (str(delay))
    if set_capacities:
        capacity = capacity_norm * topology.adj[u][v]['capacity']
        attr_str += " datarate=%sMbps;" % (str(capacity))
    %>
    ${"node[%s].ppg$o++ --> {%s} --> node[%s].pppg$i++;" % (str(node_map[u]), attr_str.strip(), str(node_map[v]))}
    % if not topology.is_directed():
    ${"node[%s].ppg$i++ <-- {%s} <-- node[%s].pppg$o++;" % (str(node_map[u]), attr_str.strip(), str(node_map[v]))}
    % endif
% endfor
}
"""

def to_omnetpp(topology, path=None):
    """Convert an FNSS topology into an Omnet++ NED script.

    Parameters
    ----------
    topology : Topology
        The topology object to convert
    path : str, optional
        The path to the output NED file.
        If not specified, prints to standard output
    """
    try:
        from mako.template import Template
    except ImportError:
        raise ImportError('Cannot import mako.template module. '
                          'Make sure mako is installed on this machine.')
    set_delays = True
    set_capacities = True
    # Check correctness of capacity and delay attributes
    if not 'capacity_unit' in topology.graph or not topology.graph['capacity_unit'] in capacity_units:
        warn('Missing or invalid capacity unit attribute in the topology. The '
             'output file will be generated without link capacity attributes.')
        set_capacities = False
    if not 'delay_unit' in topology.graph or not topology.graph['delay_unit'] in time_units:
        warn('Missing or invalid delay unit attribute in the topology. The '
             'output file will be generated without link delay attributes.')
        set_delays = False
    template = Template(__TEMPLATE)
    variables = {
        'topology':        topology,
        'set_capacities':  set_capacities,
        'set_delays':      set_delays,
                }
    ned = template.render(**variables)
    if path:
        with open(path, "w") as out:
            out.write(ned)
    else:
        print(ned)




import unittest

import networkx as nx
import fnss

class Test(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_erdos_renyi_topology(self):
        topology = fnss.erdos_renyi_topology(1000, 0.2)
        topology_fast = fnss.erdos_renyi_topology(1000, 0.2, fast=True)
        for t in (topology, topology_fast):
            self.assertEqual(1000, t.number_of_nodes())
        self.assertRaises(ValueError, fnss.erdos_renyi_topology, -1, 0.2)
        self.assertRaises(ValueError, fnss.erdos_renyi_topology, 40, 1.2)
        self.assertRaises(ValueError, fnss.erdos_renyi_topology, 50, -0.2)

    def test_erdos_renyi_topology_no_seed(self):
        a = fnss.erdos_renyi_topology(100, 0.2)
        b = fnss.erdos_renyi_topology(100, 0.2)
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_erdos_renyi_topology_constant_seed(self):
        a = fnss.erdos_renyi_topology(100, 0.2, seed=1)
        b = fnss.erdos_renyi_topology(100, 0.2, seed=2)
        c = fnss.erdos_renyi_topology(100, 0.2, seed=1)
        self.assertEqual(set(a.edges()), set(c.edges()))
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_erdos_renyi_topology_zero_seed(self):
        a = fnss.erdos_renyi_topology(100, 0.2, seed=0)
        b = fnss.erdos_renyi_topology(100, 0.2, seed=0)
        self.assertEqual(set(a.edges()), set(b.edges()))

    def test_fast_erdos_renyi_topology_no_seed(self):
        a = fnss.erdos_renyi_topology(100, 0.2, fast=True)
        b = fnss.erdos_renyi_topology(100, 0.2, fast=True)
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_fast_erdos_renyi_topology_constant_seed(self):
        a = fnss.erdos_renyi_topology(100, 0.2, fast=True, seed=1)
        b = fnss.erdos_renyi_topology(100, 0.2, fast=True, seed=2)
        c = fnss.erdos_renyi_topology(100, 0.2, fast=True, seed=1)
        self.assertEqual(set(a.edges()), set(c.edges()))
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_fast_erdos_renyi_topology_zero_seed(self):
        a = fnss.erdos_renyi_topology(100, 0.2, fast=True, seed=0)
        b = fnss.erdos_renyi_topology(100, 0.2, fast=True, seed=0)
        self.assertEqual(set(a.edges()), set(b.edges()))

    def test_waxman_1_topology(self):
        self.assertRaises(ValueError, fnss.waxman_1_topology, 0, 0.5, 0.5, 10)
        self.assertRaises(ValueError, fnss.waxman_1_topology, 10, 0, 0.5, 10)
        self.assertRaises(ValueError, fnss.waxman_1_topology, 10, 0.5, 0.5, 0)
        self.assertRaises(ValueError, fnss.waxman_1_topology, 10, 0.5, 0, 10)
        topology = fnss.waxman_1_topology(200, alpha=0.5, beta=0.6, L=20)
        self.assertEqual(200, topology.number_of_nodes())
        length = nx.get_edge_attributes(topology, 'length')
        self.assertGreaterEqual(20, max(length.values()))
        self.assertLessEqual(0, min(length.values()))

    def test_waxman_1_topology_no_seed(self):
        a = fnss.waxman_1_topology(100, alpha=0.5, beta=0.6, L=20)
        b = fnss.waxman_1_topology(100, alpha=0.5, beta=0.6, L=20)
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_waxman_1_topology_constant_seed(self):
        a = fnss.waxman_1_topology(100, alpha=0.5, beta=0.6, L=20, seed=1)
        b = fnss.waxman_1_topology(100, alpha=0.5, beta=0.6, L=20, seed=2)
        c = fnss.waxman_1_topology(100, alpha=0.5, beta=0.6, L=20, seed=1)
        self.assertEqual(set(a.edges()), set(c.edges()))
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_waxman_1_topology_zero_seed(self):
        a = fnss.waxman_1_topology(100, alpha=0.5, beta=0.6, L=20, seed=0)
        b = fnss.waxman_1_topology(100, alpha=0.5, beta=0.6, L=20, seed=0)
        self.assertEqual(set(a.edges()), set(b.edges()))

    def test_waxman_2_topology(self):
        self.assertRaises(ValueError, fnss.waxman_2_topology, 0, 0.5, 0.6)
        self.assertRaises(ValueError, fnss.waxman_2_topology, 10, 0, 0.5)
        self.assertRaises(ValueError, fnss.waxman_2_topology, 10, 0.4, 0)
        self.assertRaises(ValueError, fnss.waxman_2_topology, 10, 0.5, 0.3, (1.2, -2, 1, 2))
        self.assertRaises(ValueError, fnss.waxman_2_topology, 10, 0.5, 0.3, (-1, 3, 1, 2))
        self.assertRaises(ValueError, fnss.waxman_2_topology, 10, 0.5, 0.3, (-1, -2, 1, 2, 5))
        self.assertRaises(ValueError, fnss.waxman_2_topology, 10, 0.5, 0.3, (-1, -2, 1))
        topology = fnss.waxman_2_topology(200, alpha=0.5, beta=0.6, domain=(-1, -2, 1, 2))
        self.assertEqual(200, topology.number_of_nodes())
        # test all nodes have longitude and latitude attribute
        x = nx.get_node_attributes(topology, 'longitude')
        y = nx.get_node_attributes(topology, 'latitude')
        self.assertEqual(topology.number_of_nodes(), len(x))
        self.assertEqual(topology.number_of_nodes(), len(y))
        # test all edges have length attribute
        length = nx.get_edge_attributes(topology, 'length')
        self.assertEqual(topology.number_of_edges(), len(length))
        # test node coordinates are within predefined domain
        self.assertLessEqual(-1, min(x.values()))
        self.assertLessEqual(-2, min(y.values()))
        self.assertGreaterEqual(1, max(x.values()))
        self.assertGreaterEqual(2, max(y.values()))

    def test_waxman_2_topology_no_seed(self):
        a = fnss.waxman_2_topology(100, alpha=0.5, beta=0.6, domain=(-1, -2, 1, 2))
        b = fnss.waxman_2_topology(100, alpha=0.5, beta=0.6, domain=(-1, -2, 1, 2))
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_waxman_2_topology_constant_seed(self):
        a = fnss.waxman_2_topology(100, alpha=0.5, beta=0.6, domain=(-1, -2, 1, 2), seed=1)
        b = fnss.waxman_2_topology(100, alpha=0.5, beta=0.6, domain=(-1, -2, 1, 2), seed=2)
        c = fnss.waxman_2_topology(100, alpha=0.5, beta=0.6, domain=(-1, -2, 1, 2), seed=1)
        self.assertEqual(set(a.edges()), set(c.edges()))
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_waxman_2_topology_zero_seed(self):
        a = fnss.waxman_2_topology(100, alpha=0.5, beta=0.6, domain=(-1, -2, 1, 2), seed=0)
        b = fnss.waxman_2_topology(100, alpha=0.5, beta=0.6, domain=(-1, -2, 1, 2), seed=0)
        self.assertEqual(set(a.edges()), set(b.edges()))

    def test_barabasi_albert_topology(self):
        self.assertRaises(ValueError, fnss.barabasi_albert_topology, 0, 20, 30)
        self.assertRaises(ValueError, fnss.barabasi_albert_topology, 50, 40, 20)
        self.assertRaises(ValueError, fnss.barabasi_albert_topology, 10, 40, 20)
        self.assertRaises(ValueError, fnss.barabasi_albert_topology, 50, -1, 20)
        self.assertRaises(ValueError, fnss.barabasi_albert_topology, 10, 11, 16)
        topology = fnss.barabasi_albert_topology(100, 11, 16)
        self.assertEqual(100, topology.number_of_nodes())

    def test_barabasi_albert_topology_no_seed(self):
        a = fnss.barabasi_albert_topology(100, 11, 16)
        b = fnss.barabasi_albert_topology(100, 11, 16)
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_barabasi_albert_topology_constant_seed(self):
        a = fnss.barabasi_albert_topology(100, 11, 16, seed=1)
        b = fnss.barabasi_albert_topology(100, 11, 16, seed=2)
        c = fnss.barabasi_albert_topology(100, 11, 16, seed=1)
        self.assertEqual(set(a.edges()), set(c.edges()))
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_barabasi_albert_topology_zero_seed(self):
        a = fnss.barabasi_albert_topology(100, 11, 16, seed=0)
        b = fnss.barabasi_albert_topology(100, 11, 16, seed=0)
        self.assertEqual(set(a.edges()), set(b.edges()))

    def test_extended_barabasi_albert_topology(self):
        self.assertRaises(ValueError, fnss.extended_barabasi_albert_topology, 0, 20, 30, 0.2, 0.3)
        self.assertRaises(ValueError, fnss.extended_barabasi_albert_topology, 60, 30, 20, 0.2, 0.3)
        self.assertRaises(ValueError, fnss.extended_barabasi_albert_topology, 60, 20, 30, 0.6, 0.7)
        self.assertRaises(ValueError, fnss.extended_barabasi_albert_topology, 20, 11, 16, 1.2, 0.3)
        self.assertRaises(ValueError, fnss.extended_barabasi_albert_topology, 20, 11, 16, 0.3, 1.2)
        self.assertRaises(ValueError, fnss.extended_barabasi_albert_topology, 20, 11, 16, -0.2, 0.3)
        self.assertRaises(ValueError, fnss.extended_barabasi_albert_topology, 20, 11, 16, 0.3, -0.2)
        self.assertRaises(ValueError, fnss.extended_barabasi_albert_topology, 16, 11, 20, 0.2, 0.3)
        topology = fnss.extended_barabasi_albert_topology(100, 11, 16, 0.2, 0.3)
        self.assertEqual(100, topology.number_of_nodes())

    def test_extended_barabasi_albert_topology_no_seed(self):
        a = fnss.extended_barabasi_albert_topology(100, 11, 16, 0.2, 0.3)
        b = fnss.extended_barabasi_albert_topology(100, 11, 16, 0.2, 0.3)
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_extended_barabasi_albert_topology_constant_seed(self):
        a = fnss.extended_barabasi_albert_topology(100, 11, 16, 0.2, 0.3, seed=1)
        b = fnss.extended_barabasi_albert_topology(100, 11, 16, 0.2, 0.3, seed=2)
        c = fnss.extended_barabasi_albert_topology(100, 11, 16, 0.2, 0.3, seed=1)
        self.assertEqual(set(a.edges()), set(c.edges()))
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_extended_barabasi_albert_topology_zero_seed(self):
        a = fnss.extended_barabasi_albert_topology(100, 11, 16, 0.2, 0.3, seed=0)
        b = fnss.extended_barabasi_albert_topology(100, 11, 16, 0.2, 0.3, seed=0)
        self.assertEqual(set(a.edges()), set(b.edges()))

    def test_glp_topology(self):
        self.assertRaises(ValueError, fnss.glp_topology, 0, 20, 30, 0.4, -0.5)
        self.assertRaises(ValueError, fnss.glp_topology, 100, 30, 20, 0.4, -0.5)
        self.assertRaises(ValueError, fnss.glp_topology, 0, 30, 20, -1, -0.5)
        self.assertRaises(ValueError, fnss.glp_topology, 0, 30, 20, 0.2, -2)
        self.assertRaises(ValueError, fnss.glp_topology, 20, 11, 16, -0.5, 0.5)
        self.assertRaises(ValueError, fnss.glp_topology, 20, 11, 16, 1.5, 0.5)
        topology = fnss.glp_topology(20, 11, 16, 0.5, 0.5)
        self.assertEqual(20, topology.number_of_nodes())

    def test_glp_topology_no_seed(self):
        a = fnss.glp_topology(100, 15, 20, 0.5, 0.5)
        b = fnss.glp_topology(100, 15, 20, 0.5, 0.5)
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_glp_topology_constant_seed(self):
        a = fnss.glp_topology(100, 15, 20, 0.5, 0.5, seed=1)
        b = fnss.glp_topology(100, 15, 20, 0.5, 0.5, seed=2)
        c = fnss.glp_topology(100, 15, 20, 0.5, 0.5, seed=1)
        self.assertEqual(set(a.edges()), set(c.edges()))
        self.assertNotEqual(set(a.edges()), set(b.edges()))

    def test_glp_topology_zero_seed(self):
        a = fnss.glp_topology(100, 15, 20, 0.5, 0.5, seed=0)
        b = fnss.glp_topology(100, 15, 20, 0.5, 0.5, seed=0)
        self.assertEqual(set(a.edges()), set(b.edges()))



"""Tools for configuring network topologies

The functions provided allow users to add link capacities, delays, weights,
buffer sizes and deploy protocol stacks and applications
"""
from fnss.netconfig.buffers import *
from fnss.netconfig.capacities import *
from fnss.netconfig.delays import *
from fnss.netconfig.nodeconfig import *
from fnss.netconfig.weights import *



"""Adapter for Mininet.

This module contains function to convert FNSS topologies into Mininet
topologies and viceversa.
"""
import re

import networkx as nx

from fnss.units import capacity_units, time_units
from fnss.topologies.topology import Topology
from fnss.netconfig import set_delays_constant

__all__ = [
    'from_mininet',
    'to_mininet',
           ]

def from_mininet(topology):
    """Convert a Mininet topology to an FNSS one.

    Parameters
    ----------
    topology : Mininet Topo
        A Mininet topology object

    Returns
    -------
    topology : Topology
        An FNSS Topology object
    """
    fnss_topo = Topology(capacity_unit='Mbps')
    for v in topology.switches():
        fnss_topo.add_node(v, type='switch')
    for v in topology.hosts():
        fnss_topo.add_node(v, type='host')
    for u, v in topology.links():
        fnss_topo.add_edge(u, v)
        opts = topology.linkInfo(u, v)
        if 'bw' in opts:
            fnss_topo.adj[u][v]['capacity'] = opts['bw']
        if 'delay' in opts:
            delay = opts['delay']
            val = re.findall("\d+\.?\d*", delay)[0]
            unit = delay.strip(val).strip(' ')
            set_delays_constant(fnss_topo, val, unit, [(u, v)])
    return fnss_topo


def to_mininet(topology, switches=None, hosts=None, relabel_nodes=True):
    """Convert an FNSS topology to Mininet Topo object that can be used to
    deploy a Mininet network.

    If the links of the topology are labeled with delays, capacities or buffer
    sizes, the returned Mininet topology will also include those parameters.

    However, it should be noticed that buffer sizes are included in the
    converted topology only if they are expressed in packets. If buffer sizes
    are expressed in the form of bytes they will be discarded. This is because
    Mininet only supports buffer sizes expressed in packets.

    Parameters
    ----------
    topology : Topology, DirectedTopology or DatacenterTopology
        An FNSS Topology object
    switches : list, optional
        List of topology nodes acting as switches
    hosts : list, optional
        List of topology nodes acting as hosts
    relabel_nodes : bool, optional
        If *True*, rename node labels according to `Mininet conventions
        <https://github.com/mininet/mininet/wiki/Introduction-to-Mininet#naming-in-mininet>`_.
        In Mininet all node labels are strings whose values are "h1", "h2", ...
        if the node is a host or "s1", "s2", ... if the node is a switch.

    Returns
    -------
    topology : Mininet Topo
        A Mininet topology object

    Notes
    -----
    It is not necessary to provide a list of switch and host nodes if the
    topology object provided are already annotated with a type attribute that
    can have values *host* or *switch*. This is the case of datacenter
    topologies generated with FNSS which already include information about
    which nodes are hosts and which are switches.

    If switches and hosts are passed as arguments, then the hosts and switches
    sets must be disjoint and their union must coincide to the set of all
    topology nodes. In other words, there cannot be nodes labeled as both
    *host* and *switch* and there cannot be nodes that are neither a *host* nor
    a *switch*.

    It is important to point out that if the topology contains loops, it will
    not work with the *ovs-controller* and *controller* provided by Mininet. It
    will be necessary to use custom controllers. Further info `here
    <https://github.com/mininet/mininet/wiki/Introduction-to-Mininet#multipath-routing>`_.
    """
    try:
        from mininet.topo import Topo
    except ImportError:
        raise ImportError('Cannot import mininet.topo package. '
                          'Make sure Mininet is installed on this machine.')
    if hosts is None:
        hosts = (v for v in topology.nodes()
                 if 'host' in topology.node[v]['type'])
    if switches is None:
        switches = (v for v in topology.nodes()
                    if 'switch' in topology.node[v]['type'])
    nodes = set(topology.nodes())
    switches = set(switches)
    hosts = set(hosts)
    if not switches.isdisjoint(hosts):
        raise ValueError('Some nodes are labeled as both host and switch. '
                         'Switches and hosts node lists must be disjoint')
    if nodes != switches.union(hosts):
        raise ValueError('Some nodes are not labeled as either host or switch '
                         'or some nodes listed as switches or hosts do not '
                         'belong to the topology')
    if relabel_nodes:
        hosts = sorted(hosts)
        switches = sorted(switches)
        mapping = dict([(hosts[i], "h%s" % str(i + 1)) for i in range(len(hosts))] +
                       [(switches[i], "s%s" % str(i + 1)) for i in range(len(switches))])
        hosts = set(mapping[v] for v in hosts)
        switches = set(mapping[v] for v in switches)
        nodes = hosts.union(switches)
        topology = nx.relabel_nodes(topology, mapping, copy=True)
    topo = Topo()
    for v in switches:
        topo.addSwitch(str(v))
    for v in hosts:
        topo.addHost(str(v))
    delay_unit = topology.graph.get('delay_unit', None)
    capacity_unit = topology.graph.get('capacity_unit', None)
    buffer_unit = topology.graph.get('buffer_unit', None)
    if capacity_unit:
        capacity_conversion = float(capacity_units[capacity_unit]) \
                              / capacity_units['Mbps']
    if delay_unit:
        delay_conversion = float(time_units[delay_unit]) / time_units['us']
    for u, v in topology.edges():
        params = {}
        if 'capacity' in topology.adj[u][v] and capacity_unit:
            params['bw'] = topology.adj[u][v]['capacity'] * capacity_conversion
            # Use Token Bucket filter to implement rate limit
            params['use_htb'] = True
        if 'delay' in topology.adj[u][v] and delay_unit:
            params['delay'] = '%sus' % str(topology.adj[u][v]['delay']
                                           * delay_conversion)
        if 'buffer_size' in topology.adj[u][v] and buffer_unit == 'packets':
            params['max_queue_size'] = topology.adj[u][v]['buffer_size']
        topo.addLink(str(u), str(v), **params)
    return topo



"""
The Fast Network Simulation Setup (FNSS) core library is a Python library
providing a set of features allowing network researchers and engineers to
simplify the setup of a network simulation.

These features include the ability to:

* Parse a topology from a dataset, a topology generator or generate it
  according to a number of synthetic models.
* Apply link capacities, link weights, link delays and buffer sizes.
* Deploy protocol stacks and applications on network nodes.
* Generate traffic matrices.
* Generate event schedules.

This core library includes adapters that allow users to export topologies to
ns-2, Mininet and Autonetkit.
In addition, the core library can be used in conjunction with the FNSS Java and
C++ API or the ns-3 adapter to export topologies, traffic matrices and event
schedules to the desired target simulator or emulator.
"""
# check Python version
import sys
if sys.version_info[:2] < (2, 7):
    m = "Python version 2.7 or later is required for FNSS (%d.%d detected)."
    raise ImportError(m % sys.version_info[:2])
del sys

# Import release information
import fnss.release as release

__author__ = release.author
__version__ = release.version
__license__ = release.license_short

# import all subpackages and modules
from fnss.topologies import *
from fnss.netconfig import *
from fnss.traffic import *
from fnss.adapters import *
from fnss.units import *



import unittest

import fnss

class DatacenterTopologiesTest(unittest.TestCase):

    def test_datacenter_topology_class(self):
        topology = fnss.DatacenterTopology()
        switch_list = [1, 2 , 3]
        host_list = [4, 5, 6, 7]
        others_list = [8, 9]
        for node in switch_list:
            topology.add_node(node, type='switch')
        for node in host_list:
            topology.add_node(node, type='host')
        for node in others_list:
            topology.add_node(node, type='other')
        self.assertEquals(switch_list, topology.switches())
        self.assertEquals(host_list, topology.hosts())
        self.assertEquals(len(switch_list), topology.number_of_switches())
        self.assertEquals(len(host_list), topology.number_of_hosts())

    def test_fat_tree(self):
        topology = fnss.fat_tree_topology(8)
        self.assertEqual(len(topology), 208)

    def test_fat_tree_params(self):
        self.assertRaises(ValueError, fnss.fat_tree_topology, 0)  # zero k
        self.assertRaises(ValueError, fnss.fat_tree_topology, -1)  # negative k
        self.assertRaises(ValueError, fnss.fat_tree_topology, 11)  # odd k

    def test_two_tier(self):
        topology = fnss.two_tier_topology(10, 20, 30)
        self.assertEqual(len(topology), 10 + 20 + 20 * 30)
        self.assertEqual(topology.number_of_hosts(), 20 * 30)
        self.assertEqual(topology.number_of_switches(), 10 + 20)

    def test_two_tier_params(self):
        self.assertRaises(ValueError, fnss.two_tier_topology, 0, 10, 12)
        self.assertRaises(ValueError, fnss.two_tier_topology, 10, 0, 12)
        self.assertRaises(ValueError, fnss.two_tier_topology, 12, 10, 0)
        self.assertRaises(ValueError, fnss.two_tier_topology, -1, 10, 12)
        self.assertRaises(ValueError, fnss.two_tier_topology, 10, -1, 12)
        self.assertRaises(ValueError, fnss.two_tier_topology, 12, 10, -1)

    def test_three_tier(self):
        topology = fnss.three_tier_topology(10, 20, 5, 6)
        self.assertEqual(len(topology), 10 + 20 + 20 * 5 + 20 * 5 * 6)
        self.assertEqual(topology.number_of_hosts(), 20 * 5 * 6)
        self.assertEqual(topology.number_of_switches(), 10 + 20 + 20 * 5)

    def test_three_tier_params(self):
        self.assertRaises(ValueError, fnss.three_tier_topology, 0, 10, 12, 20)
        self.assertRaises(ValueError, fnss.three_tier_topology, 10, 0, 12, 20)
        self.assertRaises(ValueError, fnss.three_tier_topology, 12, 10, 0, 20)
        self.assertRaises(ValueError, fnss.three_tier_topology, 12, 10, 10, 0)
        self.assertRaises(ValueError, fnss.three_tier_topology, -1, 10, 12, 20)
        self.assertRaises(ValueError, fnss.three_tier_topology, 10, -1, 12, 20)
        self.assertRaises(ValueError, fnss.three_tier_topology, 12, 10, -1, 20)
        self.assertRaises(ValueError, fnss.three_tier_topology, 12, 10, 10, -1)

    def test_bcube(self):
        topology = fnss.bcube_topology(2, 3)
        self.assertEqual(topology.number_of_hosts(), 16)
        self.assertEqual(topology.number_of_switches(), 32)

    def test_bcube_params(self):
        self.assertIsNotNone(fnss.bcube_topology(1, 0))
        self.assertRaises(ValueError, fnss.bcube_topology, -1, 2)
        self.assertRaises(ValueError, fnss.bcube_topology, 0, 0)
        self.assertRaises(ValueError, fnss.bcube_topology, 0, 3)
        self.assertRaises(ValueError, fnss.bcube_topology, -1, 4)



"""Basic utility functions"""
from __future__ import division
import ast
import random
from math import pi, sqrt, sin, cos, asin

from fnss.units import EARTH_RADIUS

__all__ = [
    'split_list',
    'random_from_pdf',
    'map_func',
    'xml_cast_type',
    'xml_type',
    'xml_indent',
    'geographical_distance',
    'package_available',
          ]


def split_list(l, size):
    """Split a list into evenly sized chunks

    Parameters
    ----------
    l : list
        The list to split
    size : n
        The size of each chunk

    Returns
    -------
    A list of sub-lists

    Example
    -------
    >>> from fnss.util import split_list
    >>> split_list([1, 2, 3, 4, 5, 6], 2)
    [[1, 2], [3, 4], [5, 6]]
    """
    return [l[i: i + size] for i in range(0, len(l), size)]


def random_from_pdf(pdf, seed=None):
    """Return a random value according to a given probability density function.

    Parameters
    ----------
    pdf : dict
        Dictionary mapping values of the random variable and probability of
        occurrence. The sum of all dictionary values must be 1.
    seed : int, optional
        The seed to be used

    Returns
    -------
    key : key of pdf
        A randomly selected key from the keyset of the *pdf* parameter

    Example
    -------
    >>> pdf = {100: 0.5, 200: 0.5}
    >>> random_from_pdf(pdf)
    100 # random
    """
    # validate input parameters
    if not isinstance(pdf, dict):
        raise ValueError('The parameter pdf must be a dictionary')
    if abs(sum(pdf.values()) - 1) > 0.0001:
        raise ValueError('The sum of all probabilities must be equal to 1')
    if seed is not None:
        random.seed(seed)
    r = random.random()
    w = 0.0
    for key, value in pdf.items():
        w += value
        if r < w:
            return key


def map_func(x):
    """Execute a function with given arguments, both of them passed as an argument

    This function is used to execute map operations on a function taking an
    arbitrary number of arguments

    Parameters
    ----------
    x : tuple (func, args)
        A tuple where the first argument is the function to execute and the
        second argument is a tuple of arguments
    """
    func, args = x
    return func(*args)


def xml_cast_type(type_attrib, val):
    """Cast a value read to an XML to an appropriate Python type

    Parameters
    ----------
    type_attrib : str
        The type of the value as specified in the XML file
    val : str
        The value to cast

    Returns
    -------
    cast_val : any type
        The val argument cast to a given type
    """
    if type_attrib == 'int':
        return int(val)
    elif type_attrib == 'float':
        return float(val)
    elif type_attrib == 'boolean':
        if val == 'True':
            return True
        else:
            return False
    elif type_attrib in ('tuple', 'list', 'dict'):
        return ast.literal_eval(val)
    else:
        return val


def xml_type(val):
    """Return a type string for writing to an XML file.

    Parameters
    ----------
    val : any type
        The value

    Returns
    -------
    type : str
        The type of the value to insert in the XML file
    """
    try:
        return {str:   'string',
                int:   'int',
                bool:  'boolean',
                float: 'float',
                dict:  'dict',
                list:  'list',
                tuple: 'tuple'
        }[type(val)]
    except KeyError:
        return 'string'


def xml_indent(elem, level=0):
    """Indent the elements of the XML tree

    Parameters
    ----------
    elem : xml.etree.Element object
        XML Element to indent
    level : int, optional
        The starting indentation level
    """
    i = "\n" + (level * "  ")
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            xml_indent(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


def package_available(pkg):
    """Test whether a package is available or not

    Parameters
    ----------
    pkg : string
        Name of the package to look for

    Returns
    -------
    pkg_available : bool
        *True* if the package is available, *False* otherwise
    """
    try:
        exec('import %s' % pkg)
    except ImportError:
        return False
    else:
        return True

def geographical_distance(lat_u, lon_u, lat_v, lon_v):
    """Return geographical distance along the Earth surface between two points
    *u* and *v*

    This distance is computed using the Haversine formula.

    Parameters
    ----------
    lat_u : float
        Latitude of point *u* in degrees
    lon_u : float
        Longitude of point *u* in degrees
    lat_v : float
        Latitude of point *v* in degrees
    lon_v : float
        Longitude of point *v* in degrees

    Returns
    -------
    d : float
        The distance between *u* and *v*
    """
    lat_u = (pi / 180) * lat_u
    lon_u = (pi / 180) * lon_u
    lat_v = (pi / 180) * lat_v
    lon_v = (pi / 180) * lon_v
    return 2 * EARTH_RADIUS * asin(sqrt(sin((lat_u - lat_v) / 2) ** 2 +
                                    cos(lat_v) * cos(lat_u)
                                    * sin((lon_u - lon_v) / 2) ** 2))



"""Setup script"""
import sys
import os
from shutil import rmtree

from setuptools import setup, find_packages

# Packages required to run FNSS
requires = [
    'networkx (>=1.6)',
    'numpy (>=1.4)',
    'mako (>=0.4)'
]

# It imports release module this way because if it tried to import fnss package
# and some required dependencies were not installed, that would fail
# This is the only way to access the release module without needing all
# dependencies.
sys.path.insert(0, 'fnss')
import release
sys.path.pop(0)

# Clean tasks
if os.path.exists('MANIFEST'): os.remove('MANIFEST')
if os.path.exists('fnss.egg-info'): rmtree('fnss.egg-info')

# Main scripts
if __name__ == "__main__":
    setup(
        name='fnss',
        version=release.version,
        author=release.author,
        author_email=release.author_email,
        packages=find_packages(exclude=("test*",)),
        scripts=[
            'bin/fnss-troubleshoot',
            'bin/mn-fnss'
        ],
        url=release.url,
        download_url=release.download_url,
        license=release.license_long,
        classifiers=[
             'Development Status :: 5 - Production/Stable',
             'Intended Audience :: Developers',
             'Intended Audience :: Science/Research',
             'Intended Audience :: Telecommunications Industry',
             'License :: OSI Approved :: BSD License',
             'Natural Language :: English',
             'Operating System :: OS Independent',
             'Programming Language :: Python :: 2',
             'Programming Language :: Python :: 2.7',
             'Programming Language :: Python :: 3',
             'Programming Language :: Python :: 3.4',
             'Programming Language :: Python :: 3.5',
             'Programming Language :: Python :: 3.6',
             'Topic :: Software Development :: Libraries :: Python Modules',
             'Topic :: Scientific/Engineering',
        ],
        description=release.description_short,
        long_description=release.description_long,
        python_requires='>=2.7.9, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
        install_requires=requires,
        keywords=[
            'network',
            'simulation',
            'topology',
            'traffic matrix',
            'link capacity',
            'delay',
            'protocol stack'
        ]
    )



import unittest

import networkx as nx

import fnss

class Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # set up topology used for all traffic matrix tests
        cls.topo = fnss.k_ary_tree_topology(3, 4)
        cls.capacities = [10, 20]
        cls.odd_links = [(u, v) for (u, v) in cls.topo.edges()
                         if (u + v) % 2 == 1]
        cls.even_links = [(u, v) for (u, v) in cls.topo.edges()
                          if (u + v) % 2 == 0]
        fnss.set_capacities_random_uniform(cls.topo, cls.capacities)
        fnss.set_delays_constant(cls.topo, 3, 'ms', cls.odd_links)
        fnss.set_delays_constant(cls.topo, 12, 'ms', cls.even_links)

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        fnss.clear_weights(self.topo)

    def test_weights_constant(self):
        fnss.set_weights_constant(self.topo, 2, self.odd_links)
        fnss.set_weights_constant(self.topo, 5, self.even_links)
        self.assertTrue(all(self.topo.adj[u][v]['weight'] in [2, 5]
                            for (u, v) in self.topo.edges()))

    def test_weights_inverse_capacity(self):
        fnss.set_weights_inverse_capacity(self.topo)
        self.assertTrue(all(self.topo.adj[u][v]['weight'] in [1, 2]
                            for (u, v) in self.topo.edges()))

    def test_weights_delays(self):
        fnss.set_weights_delays(self.topo)
        self.assertTrue(all(self.topo.adj[u][v]['weight'] in [1, 4]
                            for (u, v) in self.topo.edges()))

    def test_clear_weights(self):
        # create new topology to avoid parameters pollution
        G = fnss.star_topology(12)
        fnss.set_weights_constant(G, 3, None)
        self.assertEqual(G.number_of_edges(),
                         len(nx.get_edge_attributes(G, 'weight')))
        fnss.clear_weights(G)
        self.assertEqual(0, len(nx.get_edge_attributes(G, 'weight')))



import unittest

import fnss
from fnss.util import package_available

class Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # specifying the seeds make the topology generation deterministic
        # GLP topology has been chosen because it is always connected and
        # these parameters given a topology with large diameter and variety
        # of degrees
        # 50 nodes has been chosen because eigenvector centrality tests would
        # require considerably more time
        cls.topo = fnss.glp_topology(n=50, m=1, m0=10, p=0.2, beta=-2, seed=1)
        cls.capacities = [12, 25, 489, 1091]

    def tearDown(self):
        fnss.clear_capacities(self.topo)

    def test_capacities_constant(self):
        odd_links = [(u, v) for (u, v) in self.topo.edges()
                     if (u + v) % 2 == 1]
        even_links = [(u, v) for (u, v) in self.topo.edges()
                      if (u + v) % 2 == 0]
        fnss.set_capacities_constant(self.topo, 2, 'Mbps', odd_links)
        fnss.set_capacities_constant(self.topo, 5000, 'Kbps', even_links)
        self.assertEqual('Mbps', self.topo.graph['capacity_unit'])
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in [2, 5]
                         for (u, v) in self.topo.edges()))

    def test_capacities_edge_betweenness(self):
        fnss.set_capacities_edge_betweenness(self.topo, self.capacities)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                            for (u, v) in self.topo.edges()))

    @unittest.skipUnless(package_available('scipy'), 'Requires Scipy')
    def test_capacities_edge_communicability(self):
        fnss.set_capacities_edge_communicability(self.topo, self.capacities)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                            for (u, v) in self.topo.edges()))

    @unittest.skipUnless(package_available('scipy'), 'Requires Scipy')
    def test_capacities_edge_communicability_one_capacity(self):
        fnss.set_capacities_edge_communicability(self.topo, [10])
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] == 10
                            for (u, v) in self.topo.edges()))

    def test_capacities_betweenness_gravity(self):
        fnss.set_capacities_betweenness_gravity(self.topo, self.capacities)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                            for (u, v) in self.topo.edges()))

    def test_capacities_communicability_gravity(self):
        fnss.set_capacities_communicability_gravity(self.topo, self.capacities)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                            for (u, v) in self.topo.edges()))

    def test_capacities_degree_gravity(self):
        fnss.set_capacities_degree_gravity(self.topo, self.capacities)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                            for (u, v) in self.topo.edges()))

    def test_capacities_eigenvector_gravity(self):
        fnss.set_capacities_eigenvector_gravity(self.topo, self.capacities)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                            for (u, v) in self.topo.edges()))

    def test_capacities_eigenvector_gravity_one_capacity(self):
        fnss.set_capacities_eigenvector_gravity(self.topo, [10])
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] == 10
                            for (u, v) in self.topo.edges()))

    def test_capacities_pagerank_gravity(self):
        fnss.set_capacities_pagerank_gravity(self.topo, self.capacities)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                            for (u, v) in self.topo.edges()))

    def test_capacities_random(self):
        self.assertRaises(ValueError, fnss.set_capacities_random,
                          self.topo, {10: 0.3, 20: 0.5})
        self.assertRaises(ValueError, fnss.set_capacities_random,
                          self.topo, {10: 0.3, 20: 0.8})
        fnss.set_capacities_random(self.topo, {10: 0.3, 20: 0.7})
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in (10, 20)
                            for (u, v) in self.topo.edges()))

    def test_capacities_random_uniform(self):
        fnss.set_capacities_random_uniform(self.topo, self.capacities)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                            for (u, v) in self.topo.edges()))

    def test_capacities_random_power_law(self):
        self.assertRaises(ValueError, fnss.set_capacities_random_power_law,
                      self.topo, self.capacities, alpha=0)
        self.assertRaises(ValueError, fnss.set_capacities_random_power_law,
                      self.topo, self.capacities, alpha=-0.2)
        fnss.set_capacities_random_power_law(self.topo, self.capacities)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                            for (u, v) in self.topo.edges()))

    def test_capacities_random_zipf(self):
        self.assertRaises(ValueError, fnss.set_capacities_random_zipf,
                      self.topo, self.capacities, alpha=0)
        self.assertRaises(ValueError, fnss.set_capacities_random_zipf,
                      self.topo, self.capacities, alpha=-0.2)
        fnss.set_capacities_random_zipf(self.topo, self.capacities, alpha=0.8)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                            for (u, v) in self.topo.edges()))
        fnss.clear_capacities(self.topo)
        fnss.set_capacities_random_zipf(self.topo, self.capacities, alpha=1.2)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                            for (u, v) in self.topo.edges()))

    def test_capacities_random_zipf_mandlebrot(self):
        self.assertRaises(ValueError,
                          fnss.set_capacities_random_zipf_mandelbrot,
                          self.topo, self.capacities, alpha=0)
        self.assertRaises(ValueError,
                          fnss.set_capacities_random_zipf_mandelbrot,
                          self.topo, self.capacities, alpha=-0.2)
        self.assertRaises(ValueError,
                          fnss.set_capacities_random_zipf_mandelbrot,
                          self.topo, self.capacities, alpha=0.2, q=-0.3)
        # test with alpha=0.8 and q=2.5
        fnss.set_capacities_random_zipf_mandelbrot(self.topo, self.capacities,
                                              alpha=0.8, q=2.5)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                        for (u, v) in self.topo.edges()))
        fnss.clear_capacities(self.topo)
        # test with alpha=1.2 and q=0.4
        fnss.set_capacities_random_zipf_mandelbrot(self.topo, self.capacities,
                                              alpha=1.2, q=0.4)
        self.assertTrue(all(self.topo.adj[u][v]['capacity'] in self.capacities
                        for (u, v) in self.topo.edges()))



import unittest

import fnss

class Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.topo = fnss.glp_topology(n=100, m=1, m0=10, p=0.2, beta=-2, seed=1)
        fnss.set_capacities_random_uniform(cls.topo, [10, 20, 40])
        odd_links = [(u, v) for (u, v) in cls.topo.edges()
                     if (u + v) % 2 == 1]
        even_links = [(u, v) for (u, v) in cls.topo.edges()
                      if (u + v) % 2 == 0]
        fnss.set_delays_constant(cls.topo, 2, 'ms', odd_links)
        fnss.set_delays_constant(cls.topo, 5, 'ms', even_links)
        cls.capacities = [12, 25, 489, 1091]

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        fnss.clear_buffer_sizes(self.topo)

    def test_buffer_sizes_bw_delay_prod(self):
        fnss.set_buffer_sizes_bw_delay_prod(self.topo)
        self.assertTrue(all(self.topo.adj[u][v]['buffer'] is not None
                         for (u, v) in self.topo.edges()))

    def test_buffer_sizes_bw_delay_prod_unused_links(self):
        topo = fnss.Topology()
        topo.add_edge(1, 2, weight=100)
        topo.add_edge(2, 3, weight=1)
        topo.add_edge(3, 1, weight=1)
        fnss.set_capacities_constant(topo, 10)
        fnss.set_delays_constant(topo, 2)
        fnss.set_buffer_sizes_bw_delay_prod(topo)
        self.assertTrue(all((topo.adj[u][v]['buffer'] is not None
                         for (u, v) in topo.edges())))

    def test_buffer_sizes_bw_delay_prod_unused_links_no_return_path(self):
        topo = fnss.DirectedTopology()
        topo.add_edge(1, 2, weight=100)
        topo.add_edge(1, 3, weight=1)
        topo.add_edge(3, 2, weight=1)
        fnss.set_capacities_constant(topo, 10)
        fnss.set_delays_constant(topo, 2)
        self.assertRaises(ValueError, fnss.set_buffer_sizes_bw_delay_prod, topo)

    def test_buffer_sizes_bw_delay_prod_no_return_path(self):
        topo = fnss.DirectedTopology()
        topo.add_edge(1, 2, weight=1)
        topo.add_edge(1, 3, weight=1)
        topo.add_edge(3, 2, weight=1)
        fnss.set_capacities_constant(topo, 10)
        fnss.set_delays_constant(topo, 2)
        self.assertRaises(ValueError, fnss.set_buffer_sizes_bw_delay_prod, topo)

    def test_buffers_size_link_bandwidth(self):
        fnss.set_buffer_sizes_link_bandwidth(self.topo)
        self.assertTrue(all(self.topo.adj[u][v]['buffer'] is not None
                         for (u, v) in self.topo.edges()))

    def test_buffers_size_link_bandwidth_default_size(self):
        topo = fnss.line_topology(4)
        fnss.set_capacities_constant(topo, 8, 'Mbps', [(0, 1)])
        fnss.set_capacities_constant(topo, 16, 'Mbps', [(1, 2)])
        fnss.set_buffer_sizes_link_bandwidth(topo, buffer_unit='bytes', default_size=10)
        self.assertEquals(topo.graph['buffer_unit'], 'bytes')
        self.assertEquals(topo.adj[0][1]['buffer'], 1000000)
        self.assertEquals(topo.adj[1][2]['buffer'], 2000000)
        self.assertEquals(topo.adj[2][3]['buffer'], 10)
        fnss.clear_buffer_sizes(topo)
        self.assertTrue('capacity' not in topo.adj[2][3])
        self.assertRaises(ValueError, fnss.set_buffer_sizes_link_bandwidth, topo)

    def test_buffers_size_constant(self):
        fnss.set_buffer_sizes_constant(self.topo, 65000, buffer_unit='bytes')
        self.assertTrue(all(self.topo.adj[u][v]['buffer'] == 65000
                         for (u, v) in self.topo.edges()))

    def test_buffers_size_constant_unit_mismatch(self):
        # If I try to set buffer sizes to some interfaces using a unit and some
        # other interfaces already have buffer sizes assigned using a different
        # unit, then raise an error and ask to use the unit previously used
        topo = fnss.line_topology(3)
        fnss.set_buffer_sizes_constant(topo, 10, 'packets', [(0, 1)])
        self.assertRaises(ValueError, fnss.set_buffer_sizes_constant,
                          topo, 200, 'bytes', [(1, 2)])

    def test_get_buffer_sizes(self):
        fnss.set_buffer_sizes_constant(self.topo, 65000, buffer_unit='bytes')
        buffers = fnss.get_buffer_sizes(self.topo)
        self.assertTrue(all(buffers[(u, v)] == 65000 for (u, v) in buffers))



"""Tools for exporting and importing FNSS data structures (topologies,
event schedules and traffic matrices) to/from other simulators or emulators
"""
from fnss.adapters.autonetkit import *
from fnss.adapters.mn import *
from fnss.adapters.ns2 import *
from fnss.adapters.omnetpp import *
from fnss.adapters.jfed import *



"""Adapter for AutoNetkit.

This module contains function for converting FNSS Topology objects into
NetworkX graph objects compatible with AutoNetKit and viceversa.
"""
from fnss import rename_edge_attribute


__all__ = [
    'from_autonetkit',
    'to_autonetkit'
           ]


def to_autonetkit(topology):
    """Convert an FNSS topology into a NetworkX graph object compatible for
    AutoNetKit.

    The returned graph can be saved into a GraphML file using NetworkX
    *write_graphml* function and then passed to AutoNetKit as command line
    parameter.

    The current implementation of this function only renames the weight
    attribute from *weight* to *ospf_cost*

    Parameters
    ----------
    topology : FNSS Topology
        Autonetkit topology object

    Returns
    -------
    ank_graph : FNSS topology
        an FNSS topology compatible for import to AutoNetKit
    """
    topology = topology.copy()
    rename_edge_attribute(topology, 'weight', 'ospf_cost')
    return topology


def from_autonetkit(topology):
    """Convert an AutoNetKit graph into an FNSS Topology object.

    The current implementation of this function only renames the weight
    attribute from *weight* to *ospf_cost*

    Parameters
    ----------
    topology : NetworkX graph
        An AutoNetKit NetworkX graph

    Returns
    -------
    fnss_topology : FNSS Topology
        FNSS topology
    """
    topology = topology.copy()
    rename_edge_attribute(topology, 'ospf_cost', 'weight')
    return topology



"""Function to assign and manipulate buffer sizes of network interfaces."""
import networkx as nx
from numpy import mean

from fnss.units import capacity_units, time_units


__all__ = [
    'set_buffer_sizes_bw_delay_prod',
    'set_buffer_sizes_link_bandwidth',
    'set_buffer_sizes_constant',
    'get_buffer_sizes',
    'clear_buffer_sizes',
           ]


def set_buffer_sizes_bw_delay_prod(topology, buffer_unit='bytes',
                                   packet_size=1500):
    """
    Assign a buffer sizes proportionally to the product of link bandwidth and
    average network RTT. This is a rule of thumb according to which the buffers
    of Internet routers are generally configured.

    Parameters
    ----------
    topology : Topology or DirectedTopology
        The topology on which delays are applied.
    buffer_unit : string
        The unit of buffer sizes. Supported units are: *bytes* and *packets*
    packet_size : int, optional
        The average packet size (in bytes). It used only if *packets* is
        selected as buffer size to properly calculate buffer sizes given
        bandwidth and delay values.

    Examples
    --------
    >>> import fnss
    >>> topology = fnss.erdos_renyi_topology(50, 0.2)
    >>> fnss.set_capacities_constant(topology, 10, 'Mbps')
    >>> fnss.set_delays_constant(topology, 2, 'ms')
    >>> fnss.set_buffer_sizes_bw_delay_prod(topology)
    """
    try:
        assert all(('capacity' in topology.adj[u][v]
                    for u, v in topology.edges()))
        assert all(('delay' in topology.adj[u][v]
                    for u, v in topology.edges()))
        capacity_unit = topology.graph['capacity_unit']
        delay_unit = topology.graph['delay_unit']
    except (AssertionError, KeyError):
        raise ValueError('All links must have a capacity and delay attribute')
    topology.graph['buffer_unit'] = buffer_unit
    # this filters potential self-loops which would crash the function
    edges = [(u, v) for (u, v) in topology.edges() if u != v]
    # dictionary listing all end-to-end routes in which a link appears
    route_presence = dict(zip(edges, [[] for _ in range(len(edges))]))
    # dictionary with all network routes
    route = dict(nx.all_pairs_dijkstra_path(topology, weight='weight'))
    # Dictionary storing end-to-end path delays for each OD pair
    e2e_delay = {}

    for orig in route:
        e2e_delay[orig] = {}
        for dest in route[orig]:
            path = route[orig][dest]
            if len(path) <= 1:
                continue
            path_delay = 0
            for u, v in zip(path[:-1], path[1:]):
                if 'delay' in topology.adj[u][v]:
                    if (u, v) in route_presence:
                        route_presence[(u, v)].append((orig, dest))
                    else:
                        route_presence[(v, u)].append((orig, dest))
                    path_delay += topology.adj[u][v]['delay']
                else:
                    raise ValueError('No link delays available')
            e2e_delay[orig][dest] = path_delay

    # dict containing mean RTT experienced by flows traversing a specific link
    mean_rtt_dict = {}
    for (u, v), route in route_presence.items():
        if route:
            try:
                mean_rtt = mean([e2e_delay[o][d] + e2e_delay[d][o]
                                 for (o, d) in route])
            except KeyError:
                raise ValueError('Cannot assign buffer sizes because some '
                                 'paths do not have corresponding return path')
        else:
            # if this is the case, then this link is in any shortest path,
            # not even in the one between its endpoint because there is an
            # alternative route with lower cost.
            # In this case we arbitrarily set the RTT as the RTT between the
            # link endpoint if that link was used, i.e. twice the delay of the
            # link
            if (v, u) in edges:
                mean_rtt = topology.adj[u][v]['delay'] + \
                           topology.adj[v][u]['delay']
            else:
                try:
                    mean_rtt = topology.adj[u][v]['delay'] + e2e_delay[v][u]
                except KeyError:
                    raise ValueError('Cannot assign buffer sizes because some '
                                 'paths do not have corresponding return path')
        mean_rtt_dict[(u, v)] = mean_rtt
    norm_factor = capacity_units[capacity_unit] * \
                  time_units[delay_unit] / 8000.0
    if buffer_unit == 'packets':
        norm_factor /= packet_size
    for u, v in edges:
        capacity = topology.adj[u][v]['capacity']
        buffer_size = int(mean_rtt_dict[(u, v)] * capacity * norm_factor)
        topology.adj[u][v]['buffer'] = buffer_size
    return


def set_buffer_sizes_link_bandwidth(topology, k=1.0, default_size=None,
                                    buffer_unit='bytes', packet_size=1500):
    """
    Assign a buffer sizes proportionally to the bandwidth of the interface on
    which the flush. In particularly, the buffer size will be equal to
    :math:`k \times C`, where :math:`C` is the capacity of the link in bps.

    This assignment is equal to the bandwidth-delay product if :math:`k` is the
    average RTT in seconds.

    To use this function, all links of the topology must have a *capacity*
    attribute. If the length of a link cannot be determined, it is applied the
    delay equal *default_delay* if specified, otherwise an error is returned.

    Parameters
    ----------
    topology : Topology or DirectedTopology
        The topology on which delays are applied.
    k : float, optional
        The multiplicative constant applied to capacity to derive buffer size
    default_size : float, optional
        The buffer size to be applied to interfaces whose speed is unknown. If
        it is None and at least one link does not have a capacity attribute,
        return an error
    buffer_unit : string, unit
        The unit of buffer sizes. Supported units are: *bytes* and *packets*
    packet_size : int, optional
        The average packet size (in bytes). It used only if *packets* is
        selected as buffer size to properly calculate buffer sizes given
        bandwidth and delay values.

    Examples
    --------
    >>> import fnss
    >>> topology = fnss.erdos_renyi_topology(50, 0.1)
    >>> fnss.set_capacities_constant(topology, 10, 'Mbps')
    >>> fnss.set_delays_constant(topology, 2, 'ms')
    >>> fnss.set_buffer_sizes_link_bandwidth(topology, k=1.0)
    """
    if k <= 0:
        raise ValueError('k must be a positive number')
    if default_size is None:
        if 'capacity_unit' not in topology.graph \
                or not all('capacity' in topology.adj[u][v]
                           for u, v in topology.edges()):
            raise ValueError('All links must have a capacity attribute. '
                             'Set capacity or specify a default buffer size')
    topology.graph['buffer_unit'] = buffer_unit
    # We further dived norm_factor by 8 because buffer unit is bytes
    norm_factor = capacity_units[topology.graph['capacity_unit']] / 8.0
    if buffer_unit == 'packets':
        norm_factor /= packet_size
    for u, v in topology.edges():
        if 'capacity' in topology.adj[u][v]:
            capacity = topology.adj[u][v]['capacity']
            buffer_size = int(k * capacity * norm_factor)
        else:
            buffer_size = default_size
        topology.adj[u][v]['buffer'] = buffer_size


def set_buffer_sizes_constant(topology, buffer_size, buffer_unit='bytes',
                              interfaces=None):
    """
    Assign a constant buffer size to all selected interfaces

    Parameters
    ----------
    topology : Topology or DirectedTopology
        The topology on which buffer sizes are applied.
    buffer_size : int
        The constant buffer_size to be applied to all interface
    buffer_unit : string, unit
        The unit of buffer sizes. Supported units are: *bytes* and *packets*
    interfaces : iterable container of tuples, optional
        Iterable container of selected interfaces on which buffer sizes are
        applied.
        An interface is defined by the tuple (u,v) where u is the node on which
        the interface is located and (u,v) is the link to which the buffer
        flushes.

    Examples
    --------
    >>> import fnss
    >>> topology = fnss.Topology()
    >>> topology.add_path([1, 2, 4, 5, 8])
    >>> fnss.set_buffer_sizes_constant(topology, 100000, buffer_unit='bytes', \
    ... interfaces=[(1,2), (5,8), (4,5)])
    """
    if buffer_size < 0:
        raise ValueError('The buffer_size argument cannot be negative')
    if interfaces is not None and 'buffer_unit' in topology.graph:
        curr_buffer_unit = topology.graph['buffer_unit']
        if curr_buffer_unit != buffer_unit:
            raise ValueError('The topology already contains buffer sizes ' \
                             'expressed in %s. Use that unit instead of %s' \
                             % (curr_buffer_unit, buffer_unit))
    topology.graph['buffer_unit'] = buffer_unit
    edges = topology.edges() if interfaces is None else interfaces
    for u, v in edges:
        topology.adj[u][v]['buffer'] = buffer_size


def get_buffer_sizes(topology):
    """
    Returns all the buffer sizes.

    Parameters
    ----------
    topology : Topology or DirectedTopology

    Returns
    -------
    buffer_sizes : dict
        Dictionary of buffer sizes keyed by (u, v) tuple. The key (u, v)
        represents a network interface where u is the node on which the
        interface is located and (u, v) is the link to which the buffer flushes

    Examples
    --------
    >>> import fnss
    >>> topology = fnss.Topology()
    >>> topology.add_path([1, 2, 3])
    >>> fnss.set_buffer_sizes_constant(topology, buffer_size=10)
    >>> buffer = fnss.get_buffer_sizes(topology)
    >>> buffer[(1,2)]
    10
    """
    return nx.get_edge_attributes(topology, 'buffer')


def clear_buffer_sizes(topology):
    """
    Remove all buffer sizes from the topology.

    Parameters
    ----------
    topology : Topology or DirectedTopology
        The topology whose buffer sizes are cleared
    """
    topology.graph.pop('buffer_unit', None)
    for u, v in topology.edges():
        topology.adj[u][v].pop('buffer', None)
